"""
	##########################
	Modulo scritto da
		Gianmario Salvetti
	##########################
Questa classe, contiene dei metodi che permettono di ottenere matrici per trasformazioni affini.
Le matrici sono degli operatori lineari. Vengono solitamente utilizzate per trasformare entita' come
punti (rappresentati da vettori) o gruppi di punti (rappresentati da altre matrici). I metodi che seguono
permettono di ottenere le seguenti matrici di trasformazione:\n
	- Matrice identita' \n
	- Matrice di rotazione attorno all'asse X (e inversa) \n
	- Matrice di rotazione attorno all'asse Y (e inversa) \n
	- Matrice di rotazione attorno all'asse Z (e inversa) \n
	- Matrice di rotazione attorno ad un asse \n
	- Matrice di traslazione e inversa \n
	- Matrice per la modifica di scala (e inversa) \n
	- Matrice prospettica (e inversa) \n
	- Matrice per la modifica del sistema di coordinate \n
Le matrici di questo modulo devono essere post-moltiplicate ( punto x matrice o punti x matrice)
"""
from Vectorial import Matrix
from Vectorial import *
from GeometricObjects import *
from ParticularVectors import *
from math import *

def get_identity_matrix(n):
	"""Produce la matrice identita' di dimensione n x n"""
	result = Matrix(n, n)
	for i in range(0, n):
		result[i][i] = 1.0
	return result

def get4x4Matrix(dim, m=None):
	result = get_identity_matrix(4)
	if dim < 5:
		if m != None:
			for i in range(0, dim):
				for j in range(0, dim):
					result[i][j] = m[i][j]
	return result

def __get_XY_projection_matrix(v):
	"""Genera una matrice che proietta punti sul piano xy con direzione v"""
	result = get_identity_matrix(4)
	result[2][0] = -v[0]/v[2]
	result[2][1] = -v[1]/v[2]
	result[2][2] = 0.0 #Infatti e' una matrice di proiezione
	return result

def get_XY_projection_matrix():
	result = get_identity_matrix(4)
	result[2][2] = 0.0
	return result


def get_projection_matrix(p1, p2, p3, v): #Modificare p1, p2, p3 con la classe Plane
	"""Genera una matrice che proietta punti sul piano contenente p1, p2, p3
	     con direzione di proiezione v."""
	vx = (p2-p1).getVector() #Vettore giacente sul piano p1 p2 p3
	vz = vx**((p3-p2).getVector()) # Vettore ortogonale al piano p2 p2 p3
	vy = vz**vx   # Vettore giagente sul piano p1 p2 p3
	#I tre vettori sono tra loro ortogonali

	paus1 = Point(0.0, 0.0, 0.0)
	paus2 = Point(v[0], v[1], v[2])

	m1 = get_system_reference_matrix(p1, vx, vy, vz) #1 - Cambio il sistema di coordinate
	paus1 = paus1*m1
	paus2 = paus2*m1
	m2 = __get_XY_projection_matrix( (paus2 - paus1).getVector() ) #2 - Eseguo la proiezione sul piano xy
	m3 = get_inv_matrix_system_reference(p1, vx, vy, vz)   #3 - Ripristiono il precedente sistema di coordinate

	return m1*m2*m3


def get_X_rotation_matrix(angle):
	"""Produce una matrice 4x4 di rotazione attorno all'asse X\n\n
	     1             0             0             0 \n
	     0        +cos(angle)   +sin(angle)        0 \n
	     0        -sin(angle)   +cos(angle)        0 \n
	     0             0             0             1"""
	result = get_identity_matrix(4)

	result[1][1] = cos(angle)
	result[1][2] = sin(angle)
	result[2][1] = -sin(angle)
	result[2][2] = cos(angle)

	return result

def get_inv_matrix_X_rotation(angle):
	"""Inversa della matrice di rotazione attorno all'asse x"""
	return get_X_rotation_matrix(-angle)

def get_Y_rotation_matrix(angle):
	"""Produce una matrice 4x4 di rotazione attorno all'asse Y\n\n
	+cos(angle)        0        +sin(angle)        0 \n
	     0             1             0             0\n
	-sin(angle)        0        +cos(angle)        0 \n
	     0             0             0             1"""

	result = get_identity_matrix(4)
	result[0][0] = cos(angle)
	result[0][2] = -sin(angle)
	result[2][0] = sin(angle)
	result[2][2] = cos(angle)

	return result

def get_inv_matrix_Y_rotation(angle):
	"""Inversa della matrice di rotazione attorno all'asse y"""
	return get_Y_rotation_matrix(-angle)

def get_Z_rotation_matrix(angle):
	"""Produce una matrice 4x4 di rotazione attorno all'asse Z\n\n
	+cos(angle)   +sin(angle)        0             0\n
	-sin(angle)   +cos(angle)        0             0\n
	     0             0             1             0 \n
	     0             0             0             1"""

	result = get_identity_matrix(4)
	result[0][0] = cos(angle)
	result[0][1] = sin(angle)
	result[1][0] = -sin(angle)
	result[1][1] = cos(angle)

	return result

def get_inv_matrix_Z_rotation(angle):
	"""Inversa della matrice di rotazione attorno all'asse Z."""
	return get_Z_rotation_matrix(-angle)

def get_translation_matrix(tx, ty, tz):
	"""Produce una matrice 4x4 di traslazione\n\n
	        1             0             0             0\n
		0             1             0             0\n
		0             0             1             0\n
		tx            ty            tz            1"""

	result = get_identity_matrix(4)
	result[3][0] = tx
	result[3][1] = ty
	result[3][2] = tz
	return result

def get_inv_matrix_translation(tx, ty, tz):
	"""Inversa della matrice di traslazione"""
	return get_translation_matrix(-tx, -ty, -tz)

def get_rotation_matrix(p, v, angle): #O e' meglio spin al posto di rotation?
	"""Genera un matrice di rotazione attorno ad un asse passante per il punto p e di
	   direzione v."""
	x = v.clone()
	y = get_orthogonal(x) #Prendo un vettore perpendicolare ad x
	z = x**y

	# Faccio in modo che l'asse y del sistema di riferimento conincida con l'asse di rotazione
	m1 = get_system_reference_matrix(p, x, y, z)
	# A questo punto eseguo la rotazione attorno all'asse (che ora coincide con l'asse delle ascisse)
	m2 = get_X_rotation_matrix(angle)
	# Ripristino il precedente sistema di riferimento
	m3 = get_inv_matrix_system_reference(p, x, y, z)

	return m1*m2*m3 #Eseguo il prodotto delle tre matrici

def get_scale_matrix(sx, sy, sz, p=None):
	"""Produce una matrice 4x4 di scala\n\n
	        sx            0             0             0\n
		0             sy            0             0\n
		0             0             sz            0\n
		0             0             0             1"""

	result = get_identity_matrix(4)
	result[0][0] = sx
	result[1][1] = sy
	result[2][2] = sz

	if p != None: #E' come se p assumesse il valore Point(0, 0, 0)
		result = get_translation_matrix(-p.getX(), -p.getY(), -p.getZ())*result
		result = result*get_inv_matrix_translation(-p.getX(), -p.getY(), -p.getZ())
	return result


def get_inv_matrix_scale(sx, sy, sz, p=None):
	"""Inversa della matrice di scala"""
	return get_scale_matrix(1.0/sx, 1.0/sy, 1.0/sz, p)

def get_perspective_matrix(d):
	"""Produce una matrice prospettica con punto di osservazione nell'origine
	    degli assi, e direzione -z \n\n

		1             0             0             0     \n
		0             1             0             0     \n
		0             0             0           1/d  \n
		0             0             0             1"""

	result = get_identity_matrix(4)
	result[2][2] = 0.0
	result[2][3] = 1.0/d
	return result

def get_inv_matrix_perspective(invLambda):
	"""Inversa della matrice prospettica \n
	     Modificare get_matrix_perspective prima di implementare questa funzione"""
	pass

def get_system_reference_matrix(c, vx, vy, vz):
	""" Produce una matrice che permette di modificare il sistema di riferimento degli assi:\n\n

	     	vx[0]         vy[0]         vz[0]         0\n
		vx[1]         vy[1]         vz[1]         0\n
		vx[2]         vy[2]         vz[2]         0\n
		0             0             0             1\n\n

	     I vettori vx, vy e vz devono essere tra loro ortogonali.	"""

	m1 = get_identity_matrix(4)

	vx = vx.normalize()
	vy = vy.normalize()
	vz = vz.normalize()

	for i in range(0, 3):
		m1[i][0] = vx[i]

	for i in range(0, 3):
		m1[i][1] = vy[i]

	for i in range(0, 3):
		m1[i][2] = vz[i]

	return get_inv_matrix_translation(c.getX(), c.getY(), c.getZ()) * m1

def get_inv_matrix_system_reference(c, vx, vy, vz):
	m = get_system_reference_matrix(Point(0,0,0), vx, vy, vz)
	return (m.transpose())*get_translation_matrix(c.getX(), c.getY(), c.getZ())

	"""Questa funzione produce la matrice inversa di quella prodotta dal metodo
	      get_matrix_system_reference(c, vx, vy, vz)"""
	"""
	m1 = get_identity_matrix(4)

	ps = Points(4)

	ps[0] = Point(0, 0, 0)
	ps[1] = Point(1, 0, 0)
	ps[2] = Point(0, 1, 0)
	ps[3] = Point(0, 0, 1)

	ps = ps * get_system_reference_matrix(c, vx, vy, vz)

	vx =(ps[1] - ps[0]).getVector().normalize()
	vy = (ps[2] - ps[0]).getVector().normalize()
	vz = (ps[3] - ps[0]).getVector().normalize()

	return get_system_reference_matrix(ps[0], vx, vy, vz)
	"""
