#####################################
# Gianmario Salvetti
#	file Per.py
# Modulo per lo snap ad oggetto per.
#####################################

"""
Osservazioni.
Operando su un oggetto mesh con 300 vertici, la funzione puo' apparire lenta. In effetti, alcune ottimizzazioni sul codice
possono essere fatte. Non bisogna pero' affidarsi alla sola efficienza di questa funzione. Bisogna puntare soprattutto
sull'ottimizzazione delle strutture dati. In sostanza, quando viene chiamato il metodo getSegments(), non devono essere prodotti
tutti i segmenti costituenti l'oggetto, ma dovrebbero essere prodotti solo una parte di essi (quelli contenuti in un intorno del
cursore-per esempio). Questo limiterebbe il numero di oggetti sui quali la funzione per deve operare.

"""

# Essere piu' selettivi sui moduli, in modo da rendere piu' leggibile il codice (future ottimizzazioni)
from mathutils.GeometricObjects import *
from mathutils.Vectorial import *
from mathutils.TransMatrices import *

#Potendo disporre anche di una matrice di zoom, sarebbe possibile non dover necessariamente processare tutti i punti.
#Lo zoom, permetterebbe di stabilire un valore di intorno. Il primo punto contenuto nell'intorno specificato verrebbe
#prodotto dalla funzione mid. In questo modo, il numero medio di confronti per trovare il punto desiderato, si ridurrebbe
#a n/2, dove n e' il numero di punti candidati.


""" La funzione riceve:
1 - Reference point: punto attraverso il quale deve passare la linea calcolata
2 - Cursor position: serve per selezionare la linea sulla quale calcolare la perpendicolare
3 - Oggetto: serve per ottenere i segmenti dell'oggetto


Fase 1:
Devo selezionare il segmento sul quale calcolare la perpendicolare. Inizialmente devo modificare il
sistema di riferimento, devo passare da quello globale del cursore e da quello locale dell'oggetto,
al sistema di riferimento visuale. A questo punto, elimino la coordinata z, in modo da considerare
solo la proiezione degli oggetti sul piano di visualizzazione.
"""


def per2(referencePoint, line):
	p1 = line.pa
	p2 = line.pb
	v = (p2 - p1).getVector()
	t = (v*((referencePoint - p1).getVector())) / (v*v) #Vedere dimostrazione nella documentazione
	return p1 + t*v


def per(referencePoint, cursorPosition, object):

	# Segmento selezionato
	segmentSelected = -1

	# Distanza tra il cursore e il segmento selezionato. La distanza e' intesa tra le proiezioni di
	# segmento e cursore sul piano di visualizzazione.
	distanceCursorSegmentSelected = 0

	# Segmenti che compongono l'oggetto
	segments = object.getSegments()

	# Matrici per la modifica del sistema di riferimento
	local_to_global = object.getMatrix() # Passa dal sistema locale dell'oggetto al sistema globale
	global_to_view  = object.getViewMatrix()  # Passa dal sistema locale al view system reference

	# Matrice di proiezione (serve per eliminare la coordinata z).
	XY_projection   = get_XY_projection_matrix() # Proietta gli oggetti sul piano XY

	# cp e' la posizione del cursore relativamente al piano di visualizzazione
	cP = cursorPosition*(global_to_view*XY_projection)
	# Matrice da applicare ai punti dei segmenti
	ptsMatrix = local_to_global*global_to_view*XY_projection

	# Ora seleziono il segmento piu' vicino a cursorPosition
	for i in range(0, len(segments)):
		# Trasformo i punti del segmento
		p1 = segments[i][0] * ptsMatrix
		p2 = segments[i][1] * ptsMatrix

		# Calcolo il vettore direzione del segmento
		v = (p2 - p1).getVector()

		# Calcolo il valore di t che nell'equazione parametrica del segmento mi permette di ottenere il punto
		# ottenuto dall'intersezione tra il segmento "p1 p2" e la linea perpendicolare a "p1 p2" passante per
		# "cP"
		t = (v*((cP - p1).getVector())) / (v*v)

		# Se la proiezione ortogonale di cp sulla linea contenente p1 p2 cade sul segmento p1 p2...
		if (t >= 0 and t <= 1):
			# ... ausPoint e' un punto appartenente al segmento "p1 p2"
			ausPoint = p1 + t*v # Equazione parametrica del segmento espressa in forma vettoriale
			ausDistance = abs((ausPoint - cP).getVector()) #Ottimizzare
			# Se non e' stato selezionato alcun segmento, oppure il segmento i-esimo e' piu vicino a
			# cP di quanto non sia il segmento "segmentSelected"....
			if (segmentSelected == -1 or ausDistance < distanceCursorSegmentSelected):
				# Il nuovo "segmentSelected" diventa il segmento i-esimo
				segmentSelected = i
				# La nuova distanza tra segmentSelected e cP diventa ausDistance
				distanceCursorSegmentSelected = ausDistance
	# Se e' stato selezionato un segmento...
	if segmentSelected != -1:
		# ... calcolo lo snap-intersection
		p1 = segments[segmentSelected][0]*local_to_global
		p2 = segments[segmentSelected][1]*local_to_global

		v = (p2 - p1).getVector()
		t = (v*((referencePoint - p1).getVector())) / (v*v) #Vedere dimostrazione nella documentazione

		return p1 + t*v
	else:
		return None # Se non e' stato selezionato alcun segmento ritorno l'oggetto nullo.



